// Huffman - Wyle Encode
#include "hw.h"
#include <limits.h>

#define N	256

int		l_node[N*2], r_node[N*2];
int		parent[N*2];
DATA	g_srcHW = {0};
DATA	g_dstHW = {0};
BYTE	*g_pDataDstHW = NULL;
int		bits = 0, bdata = 0;

void	inputHist(int *hist, int nn, BYTE *dataSrc, int *ctSrc);
int		getValue(int datas, BYTE *dataSrc, int *ctSrc);
int		getValueWyle(BYTE *dataSrc, int *ctSrc);
int		sgni(int x);
int		absi(int x);
void	outputHist(int *hist, int nn, DATA *data);
void	getMin2(int *hist, int hm, int *d1, int d2);
int		makeTree(int *hist, int n);
void	outputEncode(int val, int end, DATA *data);
int		getLength(int x);
void	outputWyleEncode(int val, DATA *data);

inline void inputHist(int *hist, int nn, BYTE *dataSrc, int *ctSrc)
{
	for(int i=0;i<nn;i++){
		hist[i] = getValueWyle(dataSrc, ctSrc);
	}
}// inputHist

inline int getValue(int datas, BYTE *dataSrc, int *ctSrc)
{
	int rt,bit;
	
	rt = datas - 1;
	while(rt >= N){
		bit = getBit(dataSrc, ctSrc);
		if(bit == 0){
			rt = l_node[rt];
		}
		else{
			rt = r_node[rt];
		}
	}
	return rt;
}// getValue

inline int getValueWyle(BYTE *dataSrc, int *ctSrc)
{
	int i,ct,bit,val;
	
	ct = 0;
	while(getBit(dataSrc, ctSrc) == 1){
		ct ++;
	}
	ct += 2;
	val = 0;
	for(i=0;i<ct;i++){
		bit = getBit(dataSrc, ctSrc) << i;
		val |= bit;
	}
	return val;
}// getValueWyle

bool decodeHW(BYTE *dataSrc, int *ctSrc)
{
	if(!dataSrc || !ctSrc){
		return 0;
	}
	
	int	i,size,datas,hist[N*2];
	
	size = getLong(dataSrc, ctSrc);
	
	FREE(g_pDataDstHW);
	g_pDataDstHW = (BYTE*) malloc(size);
	if(!g_pDataDstHW){
		return 0;
	}
	
	bits = 0;
	bdata = 0;
	inputHist(hist, N, dataSrc, ctSrc);
	datas = makeTree(hist, N);
	for(i=0;i<size;i++){
		g_pDataDstHW[i] = getValue(datas, dataSrc, ctSrc);
	}
	return 1;
}// decodeHW

inline int sgni(int x)
{
	if(x < 0)	return -1;
	if(x > 0)	return 1;
	return 0;
}// sgni

inline int absi(int x)
{
	if(x < 0)	return -x;
	return x;
}// absi

inline void outputHist(int *hist, int nn, DATA *data)
{
	for(int i=0;i<nn;i++){
		outputWyleEncode(hist[i], data);
	}
}// outputHist

void getMin2(int *hist, int hm, int *d1, int *d2)
{
	int i,min;
	
	min = INT_MAX - 1;
	*d1 = *d2 = -100;
	for(i=0;i<hm;i++){
		if(hist[i] < min){
			*d1 = i;
			min = hist[i];
		}
	}
	min = INT_MAX - 1;
	for(i=0;i<hm;i++){
		if(i != (*d1) && hist[i] < min){
			*d2 = i;
			min = hist[i];
		}
	}
}// getMin2

int makeTree(int *hist, int n)
{
	int hm_data;
	int i,d1,d2;
	
	hm_data = n;
	
	for(i=0;i<N*2;i++){
		l_node[i] = r_node[i] = parent[i] = 0;
	}
	
	while(1){
		getMin2(hist, hm_data, &d1, &d2);
		if(d1 < 0 || d2 < 0)	break;
		l_node[hm_data] = d1;
		r_node[hm_data] = d2;
		parent[d1] = hm_data;
		parent[d2] = -hm_data;
		hist[hm_data] = hist[d1] + hist[d2];
		hist[d1] = hist[d2] = INT_MAX;
		hm_data ++;
	}
	return hm_data;
}// makeTree

void outputEncode(int val, int end, DATA *data)
{
	int c,i,now,s1,a1;
	int code[100];
	
	c = 0;
	now = val;
	while(1){
		s1 = sgni(parent[now]);
		a1 = absi(parent[now]);
		if(s1 < 0)	code[c++] = 1;
		else		code[c++] = 0;
		if(a1 == end - 1)	break;
		now = a1;
	}
	for(i=c-1;i>=0;i--){
		putBit(code[i], data);
	}
}// outputEncode

bool encodeHW(DATA *dataDst, BYTE *dataSrc, int src_size)
{
	if(!dataDst || !dataSrc){
		return 0;
	}
	
	int i,datas,hist[N*2];
	
	for(i=0;i<N*2;i++)	hist[i] = 0;
	for(i=0;i<src_size;i++)	hist[dataSrc[i]] ++;
	putLong(src_size, dataDst);
	bits = 0;
	bdata = 0;
	outputHist(hist, N, dataDst);
	datas = makeTree(hist, N);
	for(i=0;i<src_size;i++){
		outputEncode(dataSrc[i], datas, dataDst);
	}
	flushBit(dataDst);
	return 1;
}// encodeHW

inline int getLength(int x)
{
	int len;
	
	x /= 4;
	len = 0;
	while(x > 0){
		x /= 2;
		len ++;
	}
	return len;
}// getLength

void outputWyleEncode(int val, DATA *data)
{
	int i,s,a,len;
	
	s = 1;
	a = absi(val);
	len = getLength(a);
	for(i=0;i<len;i++)	putBit(s, data);
	putBit(1-s, data);
	len += 2;
	for(i=0;i<len;i++){
		putBit(a&1, data);
		a >>= 1;
	}
}// outputWyleEncode
